import redis
import typing


class Nested(object):

    # Instance globals
    _name: str = None  # type: ignore
    _redis: redis.Redis = None  # type: ignore
    _subpath: str = None  # type: ignore

    # Type globals
    DEFAULT: typing.Any = None
    ENCODING: str = "utf-8"

    def __init__(self, name: str, redis: redis.Redis, subpath: str = "$") -> None:
        # Set internal input parameters
        self._name = name
        self._redis = redis
        self._subpath = subpath

        # Initialize the object
        self._initialize()

    @property
    def _json(self) -> typing.Any:
        return self._redis.json()  # type: ignore[no-untyped-call]

    @property
    def _absolute_name(self) -> str:
        return f".{self._name}"

    def _initialize(self) -> None:
        # Initialize a default value if required
        if not self._json.type(self._absolute_name, self._subpath):
            # Initialize sub-structure
            self._json.set(self._absolute_name, self._subpath, self.DEFAULT)

    def _fetch_value(self, subpath: str, exception: BaseException) -> typing.Any:
        # Fetch the item type
        item_type = self._json.type(self._absolute_name, subpath)

        # If the item type is None, the item is not set
        if not item_type:
            raise exception

        # Untuple item type
        item_type_value, = item_type

        # Decode the item type if needed
        if isinstance(item_type_value, bytes):
            item_type_value = item_type_value.decode(self.ENCODING)

        # Return different types as needed
        if item_type_value in NESTED_TYPES:
            # Fetch the conversion type
            nested_class, _ = NESTED_TYPES[item_type_value]

            # Convert to a nested class
            return nested_class(self._name, self._redis, subpath)

        # Fetch the item value
        item_value, = self._json.get(self._absolute_name, subpath)

        # Return item value if not a string
        if not isinstance(item_value, str):
            return item_value

        # Evaluate the values
        return eval(item_value)

    def _update_value(self, subpath: str, value: typing.Any) -> None:
        # Check if should convert to string
        if not any(isinstance(value, mapped_type) for _, mapped_type in NESTED_TYPES.values()):
            value = repr(value)

        # Set the item in the database
        self._json.set(self._absolute_name, subpath, value)

    def _delete_value(self, subpath: str, exception: BaseException) -> None:
        # Delete the item from the database
        if not self._json.delete(self._absolute_name, subpath):
            # If deletion failed, raise the exception
            raise exception


# Nested object types
NESTED_TYPES: typing.Dict[str, typing.Tuple[typing.Type[Nested], typing.Type[typing.Any]]] = dict()



import time
import string
import random
import multiprocessing

from test_dictionary import dictionary


def test_dictionary_multiprocess_rewrites(dictionary):
    # Create global things
    manager = multiprocessing.Manager()
    exceptions = manager.list()

    large_dictionary = {"".join(random.sample(list(string.ascii_letters), 10)): "".join(random.sample(list(string.ascii_letters), 10)) for _ in range(100)}

    def stress():
        for _ in range(10):
            try:
                dictionary.update(large_dictionary)
            except BaseException as e:
                # Append failure
                exceptions.append(e)

    # Create many stress processes
    processes = [multiprocessing.Process(target=stress) for _ in range(10)]

    # Execute all processes
    for p in processes:
        p.start()

    # Wait for all processes
    for p in processes:
        p.join()

    # Raise all of the exceptions
    for e in exceptions:
        raise e


def test_dictionary_kill_during_write(dictionary):
    # Create the large dictionary
    large_dictionary = {"".join(random.sample(list(string.ascii_letters), 10)): "".join(random.sample(list(string.ascii_letters), 10)) for _ in range(1000)}

    def write():
        dictionary.update(large_dictionary)

    process = multiprocessing.Process(target=write)
    process.start()

    # Sleep random amount
    time.sleep(random.random())

    # Kill the process
    process.terminate()

    # Wait for the process to stop
    process.join()

    # Check dictionary integrity
    data = dictionary.copy()

    # Make sure the dictionary was not empty
    assert data


def test_dictionary_multiprocess_kill_during_write(dictionary):
    # Create global things
    manager = multiprocessing.Manager()
    exceptions = manager.list()

    def stress():
        try:
            dictionary.update({"".join(random.sample(list(string.ascii_letters), 10)): "".join(random.sample(list(string.ascii_letters), 10)) for _ in range(100)})
        except BaseException as e:
            # Append failure
            exceptions.append(e)

    # Create many stress processes
    processes = [multiprocessing.Process(target=stress) for _ in range(10)]

    # Execute all processes
    for p in processes:
        p.start()

    # Sleep random amount
    time.sleep(random.random())

    for p in processes:
        # Kill the process
        p.terminate()

    # Wait for all processes
    for p in processes:
        p.join()

    # Check dictionary integrity
    data = dictionary.copy()

    # Make sure the dictionary was not empty
    assert data

    # Raise all of the exceptions
    for e in exceptions:
        raise e



import json
import typing
import contextlib

# Import abstract types
from collections.abc import Mapping

# Import the abstract object
from rednest.nested import Nested, NESTED_TYPES

# Create default object so that None can be used as default value
DEFAULT = object()


class Dictionary(typing.MutableMapping[str, typing.Any], Nested):

    # Bunch mode switch
    BUNCH = True

    # Type globals
    DEFAULT = {}

    def _make_subpath(self, key: str) -> str:
        # Create and return a subpath
        return f"{self._subpath}[{json.dumps(key)}]"

    def __getitem__(self, key: str) -> typing.Any:
        # Make sure key is a string
        if not isinstance(key, str):
            raise TypeError(type(key))

        # Get using subvalue
        return self._fetch_value(self._make_subpath(key), KeyError(key))

    def __setitem__(self, key: str, value: typing.Any) -> None:
        # Set using subvalue
        self._update_value(self._make_subpath(key), value)

    def __delitem__(self, key: str) -> None:
        # Delete the item from the database
        self._delete_value(self._make_subpath(key), KeyError(key))

    def __contains__(self, key: str) -> bool:  # type: ignore[override]
        # Make sure key exists in database
        return bool(self._json.type(self._absolute_name, self._make_subpath(key)))

    def __iter__(self) -> typing.Iterator[str]:
        # Fetch the object keys
        object_keys, = self._json.objkeys(self._absolute_name, self._subpath)

        # Loop over keys and decode them
        for object_key in object_keys:
            # Decode the object if needed
            if isinstance(object_key, bytes):
                object_key = object_key.decode(self.ENCODING)

            # Yield the object key
            yield object_key

    def __len__(self) -> int:
        # Fetch the object length
        object_length: typing.List[int] = self._json.objlen(self._absolute_name, self._subpath)

        # If object length is an empty list, raise a KeyError
        if not object_length:
            raise KeyError(self._subpath)

        # Untuple the result
        object_length_value, = object_length

        # Return the object length
        return object_length_value

    def __repr__(self) -> str:
        # Format the data like a dictionary
        return "{%s}" % ", ".join("%r: %r" % item for item in self.items())

    def __eq__(self, other: typing.Any) -> bool:
        # Make sure the other object is a mapping
        if not isinstance(other, Mapping):
            return False

        # Make sure all keys exist
        if set(self.keys()) != set(other.keys()):
            return False

        # Loop over all keys
        for key in self:
            # Check whether the value equals
            if self[key] != other[key]:
                return False

        # Comparison succeeded
        return True

    def pop(self, key: str, default: typing.Any = DEFAULT) -> typing.Any:
        try:
            # Fetch the original value
            value = self[key]

            # Try copying the value
            with contextlib.suppress(AttributeError):
                value = value.copy()

            # Delete the item
            del self[key]

            # Return the value
            return value
        except KeyError:
            # Check if a default is defined
            if default != DEFAULT:
                return default

            # Reraise exception
            raise

    def popitem(self) -> typing.Tuple[str, typing.Any]:
        # Convert self to list
        keys = list(self)

        # If the list is empty, raise
        if not keys:
            raise KeyError()

        # Pop a key from the list
        key = keys.pop()

        # Return the key and the value
        return key, self.pop(key)

    def copy(self) -> typing.Dict[str, typing.Any]:
        # Create initial bunch
        output = dict()

        # Loop over keys
        for key in self:
            # Fetch value of key
            value = self[key]

            # Try copying the value
            with contextlib.suppress(AttributeError):
                value = value.copy()

            # Update the bunch
            output[key] = value

        # Return the created output
        return output

    def setdefaults(self, *dictionaries: typing.Dict[str, typing.Any], **values: typing.Dict[str, typing.Any]) -> None:
        # Update values to include all dicts
        for dictionary in dictionaries:
            values.update(dictionary)

        # Loop over all items and set the default value
        for key, value in values.items():
            self.setdefault(key, value)

    # If bunch mode is enabled (on by default, define some more functions)
    if BUNCH:

        def __getattr__(self, key: str) -> typing.Any:
            try:
                return object.__getattribute__(self, key)
            except AttributeError:
                # Key is not in prototype chain, try returning
                try:
                    return self[key]
                except KeyError:
                    # Replace KeyErrors with AttributeErrors
                    raise AttributeError(key)

        def __setattr__(self, key: str, value: typing.Any) -> None:
            try:
                object.__getattribute__(self, key)
            except AttributeError:
                # Set the item
                self[key] = value
            else:
                # Key is in prototype chain, set it
                object.__setattr__(self, key, value)

        def __delattr__(self, key: str) -> None:
            try:
                object.__getattribute__(self, key)
            except AttributeError:
                # Delete the item
                try:
                    del self[key]
                except KeyError:
                    # Replace KeyErrors with AttributeErrors
                    raise AttributeError(key)
            else:
                # Key is in prototype chain, delete it
                object.__delattr__(self, key)


# Registry object type
NESTED_TYPES["object"] = (Dictionary, dict)



import pytest

from rednest import Array

from test_utilities import dictionary, array


def test_array_create(dictionary):
    # Create the list type
    dictionary["Test"] = [1, 2, 3]

    # Make sure list is created
    assert isinstance(dictionary["Test"], Array)


def test_nested_arrays(array):
    # Set subarray
    array.append([0, 1, 2, 3])

    # Make sure list is created
    assert isinstance(array[0], Array)


def test_array_append(array):
    # Set array items
    array.append(1)
    array.append(2)
    array.append(3)

    # Make sure all items are working properly
    assert array[0] == 1
    assert array[1] == 2
    assert array[2] == 3


def test_array_contains(array):
    # Set array items
    array.append(1)

    # Make sure all items are working properly
    assert 1 in array
    assert 4 not in array


def test_array_delete(array):
    # Set array items
    array.append(1)

    # Delete from list
    assert 1 in array
    del array[0]
    assert 1 not in array

    # Delete non-existent item
    with pytest.raises(IndexError):
        del array[0]


def test_array_pop(array):
    # Set array items
    array.append(1)
    array.append(2)
    array.append(3)
    array.append(4)

    # Delete from list
    assert array.pop() == 4
    assert array.pop() == 3
    assert array.pop(0) == 1
    assert len(array) == 1


def test_array_insert(array):
    # Set array items
    array.append(1)
    array.append(2)
    array.append(3)
    array.append(4)

    # Insert in list
    array.insert(0, 9)
    assert array[0] == 9

    # Insert in list
    array.insert(1, 8)
    assert array[1] == 8


def test_array_slice(array):
    # Insert to list
    for index in range(10):
        array.append(index)

    # Check slice
    assert array[3:6] == [3, 4, 5]

    # Delete slice
    del array[2:5]

    # Check slice
    assert array == [0, 1, 5, 6, 7, 8, 9]

    # Set slice
    array[3:6] = range(6)

    # Check slice
    assert array == [0, 1, 5, 0, 1, 2, 3, 4, 5, 9]

    # Get extended slices
    assert array[3:8:2] == [0, 2, 4]

    # Delete extended slices
    del array[3:8:2]

    # Check deleted slice
    assert array == [0, 1, 5, 1, 3, 5, 9]

    # Insert extended slice (should fail)
    with pytest.raises(ValueError):
        array[2:9:3] = range(10)

    # Insert extended slice
    array[2:9:3] = range(90, 92)

    # Make sure the array is as expected
    assert array == [0, 1, 90, 1, 3, 91, 9]


def test_negative_index(array):
    # Insert to list
    array.append(1)
    array.append(2)

    # Make sure the negative index works
    assert array[-1] == 2
    assert array[-2] == 1


def test_equals(array):
    # Insert to list
    array.append(1)
    array.append(2)
    array.append(3)

    # Compare
    assert array != [2, 3]
    assert array != None
    assert array != [2, 3, 4]
    assert array == [1, 2, 3]

    # This is a convenient feature, regular lists don't do this.
    assert array == (1, 2, 3)


def test_assignment(array):
    # Insert to list
    array.append(1)
    array.append(2)

    # Assign item
    array[1] = 8

    # Check assignment
    assert array == [1, 8]



import json
import typing
import functools
import contextlib

# Import abstract types
from collections.abc import Mapping, Sequence


class Encoder(json.JSONEncoder):

    def default(self, obj: typing.Any) -> typing.Any:
        # Check if the object is a keystore
        if isinstance(obj, (Mapping, Sequence)):
            # Try copying the value
            with contextlib.suppress(AttributeError):
                return obj.copy()  # type: ignore[union-attr]

        # Fallback default
        return super(Encoder, self).default(obj)


# Update the default dumps function
json.dump = functools.partial(json.dump, cls=Encoder)
json.dumps = functools.partial(json.dumps, cls=Encoder)



import pytest

from rednest import *

from test_utilities import dictionary


def test_write_read_has_delete(dictionary):
    # Make sure the dictionary does not have the item
    assert "Hello" not in dictionary

    # Write the Hello value
    dictionary["Hello"] = "World"

    # Read the Hello value
    assert dictionary["Hello"] == "World"

    # Make sure the dictionary has the Hello item
    assert "Hello" in dictionary

    # Delete the item
    del dictionary["Hello"]

    # Make sure the dictionary does not have the item
    assert "Hello" not in dictionary

    # Make sure the getter now raises
    with pytest.raises(KeyError):
        assert dictionary["Hello"] == "World"


def test_write_read_random_types(dictionary):
    # Write some random types
    dictionary["Tuple"] = (0, 1, 2, 3)
    dictionary["Bytes"] = b"AAAA\x00BBBB"
    dictionary["ByteArray"] = bytearray(range(100))

    # Make sure these types are valid now
    assert dictionary["Tuple"] == (0, 1, 2, 3)
    assert dictionary["Bytes"] == b"AAAA\x00BBBB"
    assert dictionary["ByteArray"] == bytearray(range(100))


def test_write_recursive_dicts(dictionary):
    # Write the Hello value
    dictionary["Hello"] = {"World": 42}

    # Read the Hello value
    assert dictionary["Hello"] == dict(World=42)

    # Make sure the Hello value is a dictionary
    assert isinstance(dictionary["Hello"], Dictionary)

    # Check nested bunching
    dictionary.Hello.Test = {"Value": 90}
    assert dictionary.Hello.Test.Value == 90

    # Test nesting
    assert isinstance(dictionary.Hello.Test, Dictionary)

    # Setdefault with empty dict
    dictionary.setdefault("Test", dict())

    # Make sure subdict is a dictionary
    assert isinstance(dictionary.Test, Dictionary)

    # Check type of subdict
    assert type(dictionary.Test) == Dictionary

    # Test setdefaults on subdictionary
    dictionary["Test"].setdefaults(a=1)


def test_len(dictionary):
    # Make sure dictionary is empty
    assert not dictionary

    # Load value to dictionary
    dictionary["Hello"] = "World"

    # Make sure dictionary is not empty
    assert dictionary


def test_pop(dictionary):
    # Load value to dictionary
    dictionary["Hello"] = "World"

    # Pop the item from the dictionary
    assert dictionary.pop("Hello") == "World"

    # Make sure the dictionary is empty
    assert not dictionary


def test_popitem(dictionary):
    # Load value to dictionary
    dictionary["Hello"] = "World"

    # Pop the item from the dictionary
    assert dictionary.popitem() == ("Hello", "World")

    # Make sure the dictionary is empty
    assert not dictionary


def test_copy(dictionary):
    # Load values to dictionary
    dictionary["Hello1"] = "World1"
    dictionary["Hello2"] = "World2"

    # Copy the dictionary and compare
    copy = dictionary.copy()

    # Check copy
    assert isinstance(copy, dict)
    assert copy == {"Hello1": "World1", "Hello2": "World2"}


def test_equals(dictionary):
    # Load values to dictionary
    dictionary["Hello1"] = "World1"
    dictionary["Hello2"] = "World2"

    assert dictionary == {"Hello1": "World1", "Hello2": "World2"}
    assert dictionary != {"Hello1": "World1", "Hello2": "World2", "Hello3": "World3"}
    assert dictionary != {"Hello2": "World2", "Hello3": "World3"}


def test_representation(dictionary):
    # Make sure looks good empty
    assert repr(dictionary) == "{}"

    # Load some values
    dictionary["Hello"] = "World"
    dictionary["Other"] = {"Test": 1}

    # Make sure looks good with data
    assert repr(dictionary) in ["{'Hello': 'World', 'Other': {'Test': 1}}", "{'Other': {'Test': 1}, 'Hello': 'World'}"] + ["{u'Hello': u'World', u'Other': {u'Test': 1}}", "{u'Other': {u'Test': 1}, u'Hello': u'World'}"]


def test_delete(dictionary):
    # Load some values
    dictionary["Persistent"] = "Test"
    dictionary["Volatile"] = "Forbidden"

    # Make sure values exist
    assert "Persistent" in dictionary
    assert "Volatile" in dictionary

    # Compare values
    assert dictionary["Persistent"] == "Test"
    assert dictionary["Volatile"] == "Forbidden"

    # Delete one value
    del dictionary["Volatile"]

    # Make sure persistent value exists
    assert "Persistent" in dictionary
    assert dictionary["Persistent"] == "Test"

    # Try deleting a non-existent value
    with pytest.raises(KeyError):
        del dictionary["Non-Existent"]


def test_clear(dictionary):
    # Load some values
    dictionary["Hello"] = "World"
    dictionary["Other"] = {"Test": 1}

    # Fetch other dictionary
    other = dictionary["Other"]

    # Make sure other is not empty
    assert other

    # Clear the dictionary
    dictionary.clear()

    # Make sure dictionary is empty
    assert not dictionary

    # Make sure other does not exist
    with pytest.raises(KeyError):
        assert not other


def test_setdefaults(dictionary):
    # Set some values
    dictionary.setdefaults({"a": "b"}, c="d")

    # Check dictionary structure
    assert dictionary == {"a": "b", "c": "d"}


def test_update(dictionary):
    # Set some values
    dictionary.update({"a": "b"}, c="d")

    # Check dictionary structure
    assert dictionary == {"a": "b", "c": "d"}


def test_bunch_mode(dictionary):
    # Assign values
    dictionary.test_value = 10

    # Make sure the item was written
    assert dictionary["test_value"] == 10

    # Set another value
    dictionary["another"] = 5

    # Try getting the value
    assert dictionary.another == 5

    # Try deleting the value
    del dictionary.another

    # Check final dictionary values
    assert dictionary == {"test_value": 10}



import typing
import contextlib

# Import abstract types
from collections.abc import Sequence

# Import the abstract object
from rednest.nested import Nested, NESTED_TYPES


class Array(typing.MutableSequence[typing.Any], Nested):

    # Type globals
    DEFAULT = []

    def _make_subpath(self, index: int) -> str:
        # Create and return a subpath
        return f"{self._subpath}[{index}]"

    def __getitem__(self, index: typing.Union[int, slice]) -> typing.Union[typing.Any, typing.List[typing.Any]]:
        # If a slice is provided, return a list of items
        if isinstance(index, slice):
            # Create a list with all of the requested items
            return [self[item_index] for item_index in range(*index.indices(len(self)))]
        elif isinstance(index, int):
            # Return using subvalue
            return self._fetch_value(self._make_subpath(index), IndexError(index))

        # Input type error
        raise TypeError("Array index must be an int or a slice")

    def __setitem__(self, index: typing.Union[int, slice], value: typing.Union[typing.Any, typing.Sequence[typing.Any]]) -> None:
        # If a slice is provided, splice the list
        if isinstance(index, slice):
            # Fetch slice parameters
            start, stop, step = index.indices(len(self))

            # If the step is more then 1, the value length must match the indice count
            if step > 1 and len(value) != len(range(start, stop, step)):
                raise ValueError(f"Attempted to assign sequence of incompatible size {len(value)}")

            # Create an iterator for the value
            iterator = iter(value)

            # Insert all of the items from the value
            for subindex, subvalue in zip(range(start, stop, step), iterator):
                self[subindex] = subvalue

            # If the step is exactly 1, then we want to insert new values
            if step == 1:
                # Loop over remaining items and insert them
                for counter, subvalue in enumerate(iterator):
                    self.insert(stop + counter, subvalue)
        elif isinstance(index, int):
            # Set using subvalue
            self._update_value(self._make_subpath(index), value)

    def __delitem__(self, index: typing.Union[int, slice]) -> None:
        # If a slice is provided, splice the list
        if isinstance(index, slice):
            # Delete all required items
            for counter, subindex in enumerate(range(*index.indices(len(self)))):
                del self[subindex - counter]
        elif isinstance(index, int):
            # Delete the item from the database
            self._delete_value(self._make_subpath(index), IndexError(index))

    def __len__(self) -> int:
        # Fetch the object length
        object_length: typing.List[int] = self._json.arrlen(self._absolute_name, self._subpath)

        # If object length is an empty list, raise a KeyError
        if not object_length:
            raise KeyError(self._subpath)

        # Untuple the result
        object_length_value, = object_length

        # Return the object length
        return object_length_value

    def __repr__(self) -> str:
        # Format the data like a dictionary
        return "[%s]" % ", ".join(repr(item) for item in self)

    def __eq__(self, other: typing.Any) -> bool:
        # Make sure the other item is a sequence
        if not isinstance(other, Sequence):
            return False

        # Make sure lengths are the same
        if len(self) != len(other):
            return False

        # Loop and check all items
        for index in range(len(self)):
            # Compare items
            if self[index] != other[index]:
                return False

        # Items match
        return True

    def insert(self, index: int, value: typing.Any) -> None:
        # Insert new array item
        self._json.arrinsert(self._absolute_name, self._subpath, index, value)

    def copy(self) -> typing.List[typing.Any]:
        # Create initial bunch
        output = list()

        # Loop over keys
        for value in self:
            # Try copying the value
            with contextlib.suppress(AttributeError):
                value = value.copy()

            # Update the bunch
            output.append(value)

        # Return the created output
        return output


# Registry object type
NESTED_TYPES["array"] = (Array, list)



import json
import pytest

from rednest import *

from test_utilities import dictionary, array


def test_dict_encoding(dictionary):
    # Assign values
    dictionary["AAA"] = 5
    dictionary["AAB"] = [1, 2, 3]

    # Dump dictionary
    assert json.dumps(dictionary)


def test_array_encoding(array):
    # Insert to list
    for x in range(10):
        array.append(x)

    # Encode list
    assert json.dumps(array) == '[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]'



import os
import redis
import pytest

from rednest import *

REDIS = redis.Redis()


@pytest.fixture()
def dictionary(request):
    # Generate random name
    rand_name = os.urandom(4).hex()

    # Create a random dictionary
    return Dictionary(rand_name, REDIS)


@pytest.fixture()
def array(request):
    # Generate random name
    rand_name = os.urandom(4).hex()

    # Create a random dictionary
    return Array(rand_name, REDIS)
